<?php
# @Author: Aaron Bowie
# @Date:   Friday, June 16th 2017, 11:10:13 pm
# @Email:  support@weareag.co.uk
# @Filename: epdq-class.php
# @Last modified by:   Aaron Bowie
# @Last modified time: Tuesday, November 14th 2017, 5:23:35 pm





/* Barclaycard ePDQ Direct Link Payment Gateway Class */


defined('ABSPATH') or die("No script kiddies please!");


class ag_epdq_checkout extends WC_Payment_Gateway {

	// Setup our Gateway's id, description and other values
	function __construct() {

		// The global ID for this Payment method
		$this->id = "ag_epdq_checkout";

		// The Title shown on the top of the Payment Gateways Page next to all the other Payment Gateways
		$this->method_title = __( "ePDQ Direct Link Checkout", 'ag_epdq_checkout' );

		// The description for this Payment Gateway, shown on the actual Payment options page on the backend
		$license = get_option('AGD_epdqapi_Key');
		$websiteurl = get_option ('siteurl');
		$api_params = array(
			'slm_action' => 'slm_check',
			'secret_key' => AGD_KEY,
			'license_key' => $license,
		);
		$query = esc_url_raw(add_query_arg($api_params, AGD));
		$response = wp_remote_get($query, array('timeout' => 20, 'sslverify' => false));
		$license_data = json_decode(wp_remote_retrieve_body($response));
		if($license_data->result == 'success'){
			$this->method_description = __( "Barclaycard ePDQ Direct Link Payment Gateway for WooCommerce", 'ag_epdq_checkout' );
		} else {
			$this->method_description = __( 'Please enter the license key for this product to activate it. You were given a license key when you purchased this item. <br /><strong>You can do this on the <a href="'. $websiteurl.'/wp-admin/admin.php?page=ag-epdq%2Fepdq">AG License</a> menu item on the left</strong>.', 'ag_epdq_checkout' );
		}

		// The title to be used for the vertical tabs that can be ordered top to bottom
		$this->title = __( "Barclaycard ePDQ Direct Link", 'ag_epdq_checkout' );

		// If you want to show an image next to the gateway's name on the frontend, enter a URL to an image.
		$this->icon = plugins_url('ag-epdq/new_cards.png'); // old - cards.gif

		// Bool. Can be set to true if you want payment fields to show on the checkout
		// if doing a direct integration, which we are doing in this case
		$this->has_fields = true;

		// Supports the default credit card form
		$this->supports = array( 'default_credit_card_form', 'refunds' );

		add_action('woocommerce_api_woocommerce_ag_epdq_checkout', array( $this, 'successful_request' ));
		add_action('woocommerce_receipt_ag_epdq_checkout', array($this, 'receipt_page'));
		//add_action('woocommerce_api_ag_epdq_checkout', array($this, 'check_barclay_response'));
		add_action( 'woocommerce_api_ag_epdq_checkout', array( $this, 'check_response' ) );

		$this->notify_url = WC()->api_request_url( 'ag_epdq_checkout' );


		// This basically defines your settings which are then loaded with init_settings()
		if($license_data->result == 'success'){
			$this->init_form_fields();
			$this->init_settings();
		}

		// Turn these settings into variables we can use
		foreach ( $this->settings as $setting_key => $value ) {
			$this->$setting_key = $value;
		}



		// Lets check for SSL
		add_action( 'admin_notices', array( $this,	'do_ssl_check' ) );

		// Save settings
		if ( is_admin() ) {
			// Save our administration options. Since we are not going to be doing anything special
			// we have not defined 'process_admin_options' in this class so the method in the parent
			// class will be used instead
			add_action( 'woocommerce_update_options_payment_gateways_' . $this->id, array( $this, 'process_admin_options' ) );
		}
	} // End __construct()

	// Build the administration fields for this specific Gateway
	public function init_form_fields() {
		$this->form_fields = array(
			'enabled' => array(
				'title'		=> __( 'Enable / Disable', 'ag_epdq_checkout' ),
				'label'		=> __( 'Enable this payment gateway', 'ag_epdq_checkout' ),
				'type'		=> 'checkbox',
				'default'	=> 'no',
			),
			'title' => array(
				'title'		=> __( 'Title', 'ag_epdq_checkout' ),
				'type'		=> 'text',
				'desc_tip'	=> __( 'Payment title the customer will see during the checkout process.', 'ag_epdq_checkout' ),
				'default'	=> __( 'Credit card', 'ag_epdq_checkout' ),
			),
			'description' => array(
				'title'		=> __( 'Description', 'ag_epdq_checkout' ),
				'type'		=> 'textarea',
				'desc_tip'	=> __( 'Payment description the customer will see during the checkout process.', 'ag_epdq_checkout' ),
				'default'	=> __( 'Pay securely using your credit card.', 'ag_epdq_checkout' ),
				'css'		=> 'max-width:350px;'
			),
			'api_login' => array(
				'title'		=> __( 'PSPID', 'ag_epdq_checkout' ),
				'type'		=> 'text',
				'desc_tip'	=> __( 'The PSPID for your barclays account. This is the id which you use to login the admin panel of the barclays bank.', 'ag_epdq_checkout' ),
			),
			'api_user' => array(
				'title'		=> __( 'API User ID', 'ag_epdq_checkout' ),
				'type'		=> 'text',
				'desc_tip'	=> __( '', 'ag_epdq_checkout' ),
			),
			'api_password' => array(
				'title'		=> __( 'API User Password', 'ag_epdq_checkout' ),
				'type'		=> 'text',
				'desc_tip'	=> __( '', 'ag_epdq_checkout' ),
			),
			'secure_3d' => array(
					'title' => __('Secured With 3D', 'woocommerce'),
					'type' => 'select',
					'css'  => 'height: 35px;',
					'options' => array('MAINW' => 'Main Window (Default)', 'POPUP' => 'Popup window', 'POPIX' => 'd'),
					'description' => __('Popup may be blocked by some browsers, we advicse to keep to default', 'woocommerce'),
					'default' => 'MAINW',
					'desc_tip' => false,
			),
			'operation' => array(
					'title' => __('Operation', 'woocommerce'),
					'type' => 'select',
					'css'  => 'height: 35px;',
					'options' => array('RES' => 'Authorisation', 'SAL' => 'Sale'),
					'description' => __('Operation Type For Transaction ', 'woocommerce'),
					'default' => '',//VISA;iDEA
					'desc_tip' => false,
					),

// Not needed

//      'sha_in' => array(
//          'title' => __( 'SHA-IN Passphrase', 'ag_epdq_checkout' ),
//          'type' => 'text',
//          'description' => __( 'The SHA-IN signature will encode the parameter passed to the payment processor via the hidden fields to ensure better security.', 'woocommerce' ),
//          'default' => '',
//          'desc_tip'      => true
//      ),
//      'sha_out' => array(
//          'title' => __( 'SHA-OUT Passphrase', 'ag_epdq_checkout' ),
//          'type' => 'text',
//          'description' => __( 'The SHA-OUT signature will encode the parameter passed to the redirection url from the payment processor to ensure better security.', 'woocommerce' ),
//          'default' => '',
//          'desc_tip'      => true
//      ),


//      'sha_method' => array(
//          'title' => __( 'SHA encription method', 'ag_epdq_checkout' ),
//          'type' => 'select',
//          'options'=> array(0=>'SHA-1',1=>'SHA-256',2=>'SHA-512'),
//          'description' => __( 'Sha encryption method - this needs to be similar waht you have set in the epdq backoffice.', 'woocommerce' ),
//          'default' => '',
//          'desc_tip'      => true,
//      ),


// END

			'environment' => array(
				'title'		=> __( 'Test Mode', 'ag_epdq_checkout' ),
				'label'		=> __( 'Enable Test Mode', 'ag_epdq_checkout' ),
				'type'		=> 'checkbox',
				'description' => __( 'Place the payment gateway in test mode.', 'ag_epdq_checkout' ),
				'default'	=> 'no',
			)
		);
	}

	public function credit_card_form( $args = array(), $fields = array() ) {

	  wp_enqueue_script( 'wc-credit-card-form' );

	  if( $this->environment == "yes" ) { ?>
	    <p class="form-row form-row-wide"><br /><strong>TEST MODE ACTIVE</strong><br />In test mode you can run test payments using card number <strong>4444 3333 2222 1111</strong> with any CVC number and a valid expiration date. Check our doc's for more information <a target="_blank" href="https://www.weareag.co.uk/product/barclaycard-epdq-direct-link-payment-gateway-woocommerce/">here</a>.<br /><br />If 3DSecure is enabled on the account you can test this also, the password for testing is 11111.<br /><br /></p>
	  <?php }

	  $default_args = array(
	    'fields_have_names' => true,
	  );

	  $args = wp_parse_args( $args, apply_filters( 'woocommerce_credit_card_form_args', $default_args, $this->id ) );

	  $default_fields = array(
	    'card-number-field' => '<p class="form-row form-row-wide">
	      <label for="' . esc_attr( $this->id ) . '-card-number">' . __( 'Card Number', 'woocommerce' ) . ' <span class="required">*</span></label>
	      <input id="' . esc_attr( $this->id ) . '-card-number" class="input-text wc-credit-card-form-card-number" type="text" maxlength="20" autocomplete="off" placeholder="•••• •••• •••• ••••" name="' . ( $args['fields_have_names'] ? $this->id . '-card-number' : '' ) . '" />
	    </p>',
	    'card-expiry-field' => '<p class="form-row form-row-first">
	      <label for="' . esc_attr( $this->id ) . '-card-expiry">' . __( 'Expiry (MM/YY)', 'woocommerce' ) . ' <span class="required">*</span></label>
	      <input id="' . esc_attr( $this->id ) . '-card-expiry" class="input-text wc-credit-card-form-card-expiry" type="text" autocomplete="off" placeholder="' . __( 'MM / YY', 'woocommerce' ) . '" name="' . ( $args['fields_have_names'] ? $this->id . '-card-expiry' : '' ) . '" />
	    </p>',
	    'card-cvc-field' => '<p class="form-row form-row-last">
	      <label for="' . esc_attr( $this->id ) . '-card-cvc">' . __( 'CVC', 'woocommerce' ) . ' <span class="required">*</span></label>
	      <input id="' . esc_attr( $this->id ) . '-card-cvc" class="input-text wc-credit-card-form-card-cvc" type="text" autocomplete="off" placeholder="' . __( 'CVC', 'woocommerce' ) . '" name="' . ( $args['fields_have_names'] ? $this->id . '-card-cvc' : '' ) . '" />
	    </p>'
	  );

	  $fields = wp_parse_args( $fields, apply_filters( 'woocommerce_credit_card_form_fields', $default_fields, $this->id ) );
	  ?>
	  <fieldset id="<?php echo $this->id; ?>-cc-form">
	    <?php do_action( 'woocommerce_credit_card_form_start', $this->id ); ?>
	    <?php
	      foreach ( $fields as $field ) {
	        echo $field;
	      }
	    ?>
	    <?php do_action( 'woocommerce_credit_card_form_end', $this->id ); ?>
	    <div class="clear"></div>
	  </fieldset>
	  <?php
	}


	public function check_response() {
			ob_clean();
			header('HTTP/1.1 200 OK');

			$datacheck1 = array();

			foreach ($_REQUEST as $key => $value) {
					if ($value == "") {
							continue;
					}
					$datacheck1[strtoupper($key)] = strtoupper($value);
			}

				$this->transaction_successfull($datacheck1);

	}


	function transaction_successfull($args)
	{
			global $woocommerce;

			extract($args);
			$order = new WC_Order($ORDERID);

			$accepted = array(4, 5, 9, 41, 51, 91);
			$status = $STATUS;
			$ncerror = $NCERROR;

			$dienote = '<p>Transection result is uncertain.<p>';
			$dienote .= '<p>Status Code: ' . $STATUS . ' - ' . get_epdq_status_code($status) . '';
			$dienote .= '<br>Error Code: ' . $NCERROR . ' - ' . get_epdq_ncerror($ncerror) . '</p>';
			$died = '';
			$died .= $dienote;
			$died .= '<p>Your order is cancelled and your cart is emptied.';
			$died .= '<br>Go to your <a href="' . get_permalink(get_option('woocommerce_myaccount_page_id')) . '">account</a> to process your order again or ';
			$died .= 'go to <a href="' . home_url() . '">homepage</a></p>';

			if (in_array($STATUS, $accepted)) {

					if (!empty($args['ORDERID']))
							$note = 'Order ID: ' . $ORDERID . '.<br>'; //	order id
					if (!empty($args['AMOUNT']))
							$note .= 'Amount: ' . $AMOUNT . '.<br>'; //	amount
					if (!empty($args['CURRENCY']))
							$note .= 'Order currency: ' . $CURRENCY . '.<br>'; //	order currency
					if (!empty($args['PM']))
							$note .= 'Payment Method: ' . $PM . '.<br>'; //	payment method
					if (!empty($args['ACCEPTANCE']))
							$note .= 'Acceptance code returned by acquirer: ' . $ACCEPTANCE . '.<br>';    //	acceptance
					if (!empty($args['STATUS']))
							$note .= 'Transaction status : ' . $STATUS . '.<br>'; //	status code
					//if (!empty($args['CARDNO']))
					//		$note .= 'Masked card number : ' . $CARDNO . '.<br>'; //	catd no
					if (!empty($args['PAYID']))
							$note .= 'Payment reference in EPDQ system: ' . $PAYID . '.<br>'; //	pay id
					if (!empty($args['NCERROR']))
							$note .= 'Error Code: ' . $NCERROR . '.<br>'; //	ncerror
					if (!empty($args['BRAND']))
							$note .= 'Card brand (EPDQ system derives this from the card number) : ' . $BRAND . '.<br>'; //	brand
					//if (!empty($args['ED']))
							//$note .= 'Payer\'s card expiry date : ' . $ED . '.<br>'; //	expiry date
					if (!empty($args['TRXDATE']))
							$note .= 'Transaction Date: ' . $TRXDATE . '.<br>'; //	date
					if (!empty($args['CN']))
							$note .= 'Cardholder/customer name: ' . $CN . '.<br>'; //	payer's name
					if (!empty($args['IP']))
							$note .= 'Customer\'s IP: ' . $IP . '.<br>'; //	payer's ip

					if (!empty($args['AAVADDRESS']))
							$note .= 'AAV result for the address: ' . $AAVADDRESS . ' . <br>'; //	aav address
					if (!empty($args['AAVCHECK']))
							$note .= 'Result of the automatic address verification: ' . $AAVCHECK . ' . <br>'; //	aav check
					if (!empty($args['AAVZIP']))
							$note .= 'AAV result for the zip code: ' . $AAVZIP . ' . <br>'; // aav zip
					//if (!empty($args['BIN']))
							//$note .= 'First 6 digits of credit card number: ' . $BIN . ' . <br>'; // bin
					if (!empty($args['CCCTY']))
							$note .= 'Country where the card was issued: ' . $CCCTY . ' . <br>';
					if (!empty($args['COMPLUS']))
							$note .= 'Custom value passed: ' . $COMPLUS . ' . <br>';

					if (!empty($args['CVCCHECK']))
							$note .= 'Result of the card verification code check: ' . $CVCCHECK . ' . <br>';
					if (!empty($args['ECI']))
							$note .= 'Electronic Commerce Indicator: ' . $ECI . ' . <br>';
					if (!empty($args['FXAMOUNT']))
							$note .= 'FXAMOUNT: ' . $FXAMOUNT . ' . <br>';
					if (!empty($args['FXCURRENCY']))
							$note .= 'FXCURRENCY: ' . $FXCURRENCY . ' . <br>';
					if (!empty($args['IPCTY']))
							$note .= 'Originating country of the IP address: ' . $IPCTY . ' . <br>';
					if (!empty($args['SUBBRAND']))
							$note .= 'SUBBRAND: ' . $SUBBRAND . ' . <br>';
					if (!empty($args['VC']))
							$note .= 'Virtual Card type: ' . $SUBBRAND . ' . <br>';

							// Store these for later
							if (isset($args['ORDERID'])) {
								update_post_meta($order->id, 'OrderID', $args['ORDERID']);
							}
							if (isset($args['AMOUNT'])) {
								update_post_meta($order->id, 'OrderAmount', $args['AMOUNT']);
							}
							if (isset($args['CURRENCY'])) {
								update_post_meta($order->id, 'OrderCurrency', $args['CURRENCY']);
							}
							if (isset($args['PM'])) {
								update_post_meta($order->id, 'PaymentMethod', $args['PM']);
							}
							if (isset($args['ACCEPTANCE'])) {
								update_post_meta($order->id, 'Acceptance', $args['ACCEPTANCE']);
							}
							if (isset($args['STATUS'])) {
								update_post_meta($order->id, 'Status', $args['STATUS']);
							}
							if (isset($args['PAYID'])) {
								update_post_meta($order->id, 'PAYID', $args['PAYID']);
							}
							if (isset($args['NCERROR'])) {
								update_post_meta($order->id, 'NCERROR', $args['NCERROR']);
							}
							if (isset($args['BRAND'])) {
								update_post_meta($order->id, 'BRAND', $args['BRAND']);
							}
							if (isset($args['TRXDATE'])) {
								update_post_meta($order->id, 'TRXDATE', $args['TRXDATE']);
							}
							if (isset($args['IP'])) {
								update_post_meta($order->id, 'IP', $args['IP']);
							}



					$woocommerce->cart->empty_cart();

					if (in_array($STATUS, array(4, 5, 9))) {
						$noteTitle = 'Barclay ePDQ transaction is confirmed.';
						$order->add_order_note($note);
						$order->add_order_note($noteTitle);
						$order->payment_complete();
					}

					if (in_array($STATUS, array(41, 51, 91))) {
							$note = 'Barclay ePDQ transaction is awaiting for confirmation.<br>';
							$note .= $note;
							$order->update_status('on-hold', $note);
					}

			} elseif ($STATUS == 2 || $STATUS == 93) {
					$dienote .= '<br>Order is failed.';
					$order->update_status('failed', $dienote);
					$woocommerce->cart->empty_cart();
			} elseif ($STATUS == 52 || $STATUS == 92) {
					$dienote .= '<br>Order is failed.';
					$order->update_status('failed', $dienote);
					$woocommerce->cart->empty_cart();
			} elseif ($STATUS == 1) {
					$dienote .= '<br>Order is cancelled.';
					$order->update_status('cancelled', $dienote);
					$woocommerce->cart->empty_cart();
			} else {
					$dienote .= '<br>Order is failed.';
					$order->update_status('failed', $dienote);
					$woocommerce->cart->empty_cart();
			}

			wp_redirect($this->get_return_url($order));
			exit;
	}

	public function receipt_page($order) {
			global $woocommerce;
			echo '<p>' . __('Thank you for your order, please click the button below to pay with Barclay ePDQ.', 'ag_epdq_checkout') . '</p>';
			wc_enqueue_js('
			jQuery("body").block({
				message: "<img src=\"'.esc_url($woocommerce->plugin_url()).'/assets/images/ajax-loader.gif\" alt=\"Redirecting...\" style=\"float:left; margin-right: 10px;\" />'.__('Thank you for your order. We are now redirecting you to verify your card.', 'ag_epdq_checkout').'",
				overlayCSS:
				{
					background: "#fff",
					opacity: 0.6
				},
				css:
				{
					padding:        20,
					textAlign:      "center",
					color:          "#555",
					border:         "3px solid #aaa",
					backgroundColor:"#fff",
					cursor:         "wait",
					lineHeight:		"32px"
				}
			});
			jQuery("#submit_ag_epdq_checkout_payment_form").click();
	');

			$str = WC()->session->get('HTML_ANSWER');
			$html = base64_decode($str);
			print_r($html);
	}



	// Submit payment and handle response
	public function process_payment( $order_id ) {

		global $woocommerce;

		// Get this Order's information so that we know
		// who to charge and how much
		$customer_order = new WC_Order( $order_id );

		// Are we testing right now or is it a real transaction
		$environment = ( $this->environment == "yes" ) ? 'TRUE' : 'FALSE';

		// Decide which URL to post to
		if( is_ssl()) {
			$environment_url = ( "FALSE" == $environment )
							   ? 'https://payments.epdq.co.uk/ncol/prod/orderdirect.asp'
							   : 'https://mdepayments.epdq.co.uk/ncol/test/orderdirect.asp';
		} else {
			$environment_url = 'https://mdepayments.epdq.co.uk/ncol/test/orderdirect.asp';
		}


		// This is where the fun stuff begins
    $Passphrase = $this->sha_in; // your SHA-IN pass phrase goes here
    $data_string = '';

    foreach ($data as $key => $value) {
      $data_string[] = $key. '='.$value.$Passphrase; // creates the string to hash
    }
    $string_post = '';
    $string_post = implode ('', $data_string);
    $SHASign = sha1 ($string_post);

		// sorting field
		$fullname = $customer_order->billing_first_name . ' - ' . $customer_order->billing_last_name;
		$amount = floatval( preg_replace( '#[^\d.]#', '', $customer_order->order_total*100 ) );
		$order_item = $customer_order->get_items();
		foreach( $order_item as $product ) {
		  $prodct_name[] = $product['name']." x".$product['qty'];
		}
		$product_list = implode( ',', $prodct_name );

    //construct the array again and post the SHASign as a name value pair
    $data_post;
    $data_post['PSPID'] = $this->api_login;
    $data_post['ORDERID'] = $customer_order->get_order_number();
    $data_post['USERID'] = $this->api_user;
    $data_post['PSWD'] = $this->api_password;
    $data_post['AMOUNT'] = $amount;
    $data_post['CURRENCY'] = get_woocommerce_currency();
    $data_post['CARDNO'] = str_replace( array(' ', '-' ), '', $_POST['ag_epdq_checkout-card-number'] );
    $data_post['ED'] = str_replace( array( '/', ' '), '', $_POST['ag_epdq_checkout-card-expiry'] );
    $data_post['CVC'] = ( isset( $_POST['ag_epdq_checkout-card-cvc'] ) ) ? $_POST['ag_epdq_checkout-card-cvc'] : '';
    $data_post['SHASign'] = $SHASign;
		$data_post['EMAIL'] = $customer_order->billing_email;
		$data_post['CN'] = $fullname;
		$data_post['COM'] = $product_list;
		$data_post['OWNERADDRESS'] = $customer_order->billing_address_1;
		$data_post['OWNERZIP'] = $customer_order->billing_postcode;
		$data_post['OWNERCTY'] = $customer_order->billing_city;
		$data_post['OWNERTELNO'] = $customer_order->billing_phone;
		$data_post['REMOTE_ADDR'] = $_SERVER['REMOTE_ADDR'];
		// 3D secure bits
		$data_post['WIN3DS'] = $this->secure_3d;
		$data_post['FLAG3D'] = 'Y';
		$data_post['ACCEPTURL'] = $this->notify_url;
		$data_post['DECLINEURL'] = $this->notify_url;
		$data_post['EXCEPTIONURL'] = $this->notify_url;
		$data_post['LANGUAGE'] = get_bloginfo('language');
		$data_post['OPERATION'] = $this->operation;

    $post_string = '';

    foreach ($data_post as $key => $value) {
      $post_string[] = $key. '=' .$value;
    }

    $actual_string = '';
    $actual_string = implode ('&', $post_string);
		$epdq_connection = curl_init($environment_url);

		curl_setopt ($epdq_connection, CURLOPT_CONNECTTIMEOUT, 20);
		curl_setopt ($epdq_connection, CURLOPT_RETURNTRANSFER, true);
		curl_setopt ($epdq_connection, CURLOPT_SSL_VERIFYPEER, false);
		curl_setopt($epdq_connection,CURLOPT_USERAGENT,'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US; rv:1.8.1.13) Gecko/20080311 Firefox/2.0.0.13');
		curl_setopt ($epdq_connection, CURLOPT_POSTFIELDS, $actual_string);

		$result = curl_exec($epdq_connection);
		$http_code = curl_getinfo($epdq_connection, CURLINFO_HTTP_CODE);
		$xmldata = XMLtoArray($result);
		//$this->transaction_successfull($xmldata);
		curl_close($epdq_connection);

		// Test the code to know if the transaction went through or not.
		if ( ( $xmldata['NCRESPONSE']['STATUS'] == 5 ) || ( $xmldata['NCRESPONSE']['STATUS'] == 9 ) ) {

			if (!empty($xmldata['NCRESPONSE']['ORDERID']))
					$note = 'Order ID: ' . $xmldata['NCRESPONSE']['ORDERID'] . '.<br>'; //	order id
			if (!empty($xmldata['NCRESPONSE']['AMOUNT']))
					$note .= 'Amount: ' . $xmldata['NCRESPONSE']['AMOUNT'] . '.<br>'; //	amount
			if (!empty($xmldata['NCRESPONSE']['CURRENCY']))
					$note .= 'Order currency: ' . $xmldata['NCRESPONSE']['CURRENCY'] . '.<br>'; //	order currency
			if (!empty($xmldata['NCRESPONSE']['PM']))
					$note .= 'Payment Method: ' . $xmldata['NCRESPONSE']['PM'] . '.<br>'; //	payment method
			if (!empty($xmldata['NCRESPONSE']['ACCEPTANCE']))
					$note .= 'Acceptance code returned by acquirer: ' . $xmldata['NCRESPONSE']['ACCEPTANCE'] . '.<br>';    //	acceptance
			if (!empty($xmldata['NCRESPONSE']['STATUS']))
					$note .= 'Transaction status : ' . get_epdq_status_code($xmldata['NCRESPONSE']['STATUS']) . '.<br>'; //	status code
			//if (!empty($args['CARDNO']))
			//		$note .= 'Masked card number : ' . $CARDNO . '.<br>'; //	catd no
			if (!empty($xmldata['NCRESPONSE']['PAYID']))
					$note .= 'Payment reference in ePDQ system: ' . $xmldata['NCRESPONSE']['PAYID'] . '.<br>'; //	pay id
			if (!empty($xmldata['NCRESPONSE']['NCERROR']))
					$note .= 'Error Code: ' . $xmldata['NCRESPONSE']['NCERROR'] . '.<br>'; //	ncerror
			if (!empty($xmldata['NCRESPONSE']['BRAND']))
					$note .= 'Card brand (ePDQ system derives this from the card number) : ' . $xmldata['NCRESPONSE']['BRAND'] . '.<br>'; //	brand
			//if (!empty($args['ED']))
					//$note .= 'Payer\'s card expiry date : ' . $ED . '.<br>'; //	expiry date
			if (!empty($xmldata['NCRESPONSE']['TRXDATE']))
					$note .= 'Transaction Date: ' . $xmldata['NCRESPONSE']['TRXDATE'] . '.<br>'; //	date
			if (!empty($xmldata['NCRESPONSE']['CN']))
					$note .= 'Cardholder/customer name: ' . $xmldata['NCRESPONSE']['CN'] . '.<br>'; //	payer's name
			if (!empty($xmldata['NCRESPONSE']['IP']))
					$note .= 'Customer\'s IP: ' . $xmldata['NCRESPONSE']['IP'] . '.<br>'; //	payer's ip

			if (!empty($xmldata['NCRESPONSE']['AAVADDRESS']))
					$note .= 'AAV result for the address: ' . $xmldata['NCRESPONSE']['AAVADDRESS'] . ' . <br>'; //	aav address
			if (!empty($xmldata['NCRESPONSE']['AAVCHECK']))
					$note .= 'Result of the automatic address verification: ' . $xmldata['NCRESPONSE']['AAVCHECK'] . ' . <br>'; //	aav check
			if (!empty($xmldata['NCRESPONSE']['AAVZIP']))
					$note .= 'AAV result for the zip code: ' . $xmldata['NCRESPONSE']['AAVZIP'] . ' . <br>'; // aav zip
			//if (!empty($args['BIN']))
					//$note .= 'First 6 digits of credit card number: ' . $BIN . ' . <br>'; // bin
			if (!empty($xmldata['NCRESPONSE']['CCCTY']))
					$note .= 'Country where the card was issued: ' . $xmldata['NCRESPONSE']['CCCTY'] . ' . <br>';
			if (!empty($xmldata['NCRESPONSE']['COMPLUS']))
					$note .= 'Custom value passed: ' . $xmldata['NCRESPONSE']['COMPLUS'] . ' . <br>';

			if (!empty($xmldata['NCRESPONSE']['CVCCHECK']))
					$note .= 'Result of the card verification code check: ' . $xmldata['NCRESPONSE']['CVCCHECK'] . ' . <br>';
			if (!empty($xmldata['NCRESPONSE']['ECI']))
					$note .= 'Electronic Commerce Indicator: ' . $xmldata['NCRESPONSE']['ECI'] . ' . <br>';
			if (!empty($xmldata['NCRESPONSE']['FXAMOUNT']))
					$note .= 'FXAMOUNT: ' . $xmldata['NCRESPONSE']['FXAMOUNT'] . ' . <br>';
			if (!empty($xmldata['NCRESPONSE']['FXCURRENCY']))
					$note .= 'FXCURRENCY: ' . $xmldata['NCRESPONSE']['FXCURRENCY'] . ' . <br>';
			if (!empty($xmldata['NCRESPONSE']['IPCTY']))
					$note .= 'Originating country of the IP address: ' . $xmldata['NCRESPONSE']['IPCTY'] . ' . <br>';
			if (!empty($xmldata['NCRESPONSE']['SUBBRAND']))
					$note .= 'SUBBRAND: ' . $xmldata['NCRESPONSE']['SUBBRAND'] . ' . <br>';
			if (!empty($xmldata['NCRESPONSE']['VC']))
					$note .= 'Virtual Card type: ' . $xmldata['NCRESPONSE']['VC'] . ' . <br>';

					// Store these for later
					if (isset($xmldata['NCRESPONSE']['ORDERID'])) {
						update_post_meta($customer_order->id, 'OrderID', $xmldata['NCRESPONSE']['ORDERID']);
					}
					if (isset($xmldata['NCRESPONSE']['AMOUNT'])) {
						update_post_meta($customer_order->id, 'OrderAmount', $xmldata['NCRESPONSE']['AMOUNT']);
					}
					if (isset($xmldata['NCRESPONSE']['CURRENCY'])) {
						update_post_meta($customer_order->id, 'OrderCurrency', $xmldata['NCRESPONSE']['CURRENCY']);
					}
					if (isset($xmldata['NCRESPONSE']['PM'])) {
						update_post_meta($customer_order->id, 'PaymentMethod', $xmldata['NCRESPONSE']['PM']);
					}
					if (isset($xmldata['NCRESPONSE']['ACCEPTANCE'])) {
						update_post_meta($customer_order->id, 'Acceptance', $xmldata['NCRESPONSE']['ACCEPTANCE']);
					}
					if (isset($xmldata['NCRESPONSE']['STATUS'])) {
						update_post_meta($customer_order->id, 'Status', get_epdq_status_code($xmldata['NCRESPONSE']['STATUS']));
					}
					if (isset($xmldata['NCRESPONSE']['PAYID'])) {
						update_post_meta($customer_order->id, 'PAYID', $xmldata['NCRESPONSE']['PAYID']);
					}
					if (isset($xmldata['NCRESPONSE']['NCERROR'])) {
						update_post_meta($customer_order->id, 'NCERROR', $xmldata['NCRESPONSE']['NCERROR']);
					}
					if (isset($xmldata['NCRESPONSE']['BRAND'])) {
						update_post_meta($customer_order->id, 'BRAND', $xmldata['NCRESPONSE']['BRAND']);
					}
					if (isset($xmldata['NCRESPONSE']['TRXDATE'])) {
						update_post_meta($customer_order->id, 'TRXDATE', $xmldata['NCRESPONSE']['TRXDATE']);
					}
					if (isset($xmldata['NCRESPONSE']['IP'])) {
						update_post_meta($customer_order->id, 'IP', $xmldata['NCRESPONSE']['IP']);
					}

			// Payment has been successful
			$customer_order->add_order_note( __( 'ePDQ payment completed.', 'ag_epdq_checkout' ) );

			// Mark order as Paid
			$customer_order->payment_complete();

			// Empty the cart (Very important step)
			$woocommerce->cart->empty_cart();

			// Add status code & ePDQ order ID to order
			//$customer_order->add_order_note( 'ePDQ order id: '. $xmldata['NCRESPONSE']['PAYID'] );
			//$customer_order->add_order_note( 'Status Code: '. get_epdq_status_code($xmldata['NCRESPONSE']['STATUS']) );
			$customer_order->add_order_note($note);

			// Redirect to thank you page
			return array(
				'result'   => 'success',
				'redirect' => $this->get_return_url( $customer_order ),
			);

		} elseif ( ( $xmldata['NCRESPONSE']['STATUS'] == 46 ) ) {

			WC()->session->set('HTML_ANSWER', $xmldata['NCRESPONSE']['HTML_ANSWER']);
			WC()->session->set('PaReq', $xmldata['NCRESPONSE']['PaReq']);
			WC()->session->set('Order_ID', $order_id);

			$redirect = $customer_order->get_checkout_payment_url(true);

			return array(
					'result'    => 'success',
					'redirect'    => $redirect
			);


		} else {
			// Transaction was not succesful
			// Add notice to the cart
			//wc_add_notice( get_epdq_ncerror($xmldata['NCRESPONSE']['NCERROR']), 'error' );
			wc_add_notice( get_epdq_status_code($xmldata['NCRESPONSE']['STATUS']), 'error' );

			// Add note to the order for your reference
			$customer_order->add_order_note( 'Error: '. get_epdq_ncerror($xmldata['NCRESPONSE']['NCERROR']) );
			$customer_order->add_order_note( 'Status Code: '. get_epdq_status_code($xmldata['NCRESPONSE']['STATUS']) );
		}

	}

	// Validate fields
	public function validate_fields() {
		return true;
	}

	// Check if we are forcing SSL on checkout pages
	// Custom function not required by the Gateway
	public function do_ssl_check() {
		//if( $this->enabled == "yes" ) {
			if( is_ssl() == false ) {
				echo "<div class=\"error\"><p>". sprintf( __( "<strong>%s</strong> is enabled, but you dont have an SSL certificate on your website. Please ensure that you have a valid SSL certificate.<br /><strong>ePDQ Direct Link will only work in test mode while there is no SSL</strong>" ), $this->method_title, admin_url( 'admin.php?page=wc-settings&tab=checkout' ) ) ."</p></div>";
			}
		//}
	}

	public function process_refund( $order_id, $amount = NULL, $reason = '' ) {

				$order 			   = new WC_Order( $order_id );

				if( is_ssl()) {
					$environment_url = ( "FALSE" == $environment )
										 ? 'https://mdepayments.epdq.co.uk/ncol/prod/maintenancedirect.asp'
										 : 'https://mdepayments.epdq.co.uk/ncol/test/maintenancedirect.asp';
				} else {
					$environment_url = 'https://mdepayments.epdq.co.uk/ncol/test/maintenancedirect.asp';
				}

				$Refundamount = floatval( preg_replace( '#[^\d.]#', '', $amount*100 ) );

				$data_post;
				$data_post['AMOUNT'] = $Refundamount;
				$data_post['OPERATION'] = 'RFD';
				$data_post['PAYID'] = get_post_meta( $order_id, 'PAYID', true );
				$data_post['PSPID'] = $this->api_login;
				$data_post['PSWD'] = $this->api_password;
				$data_post['USERID'] = $this->api_user;

				$post_string = '';

				foreach ($data_post as $key => $value) {
				  $post_string[] = $key. '=' .$value;
				}
				$actual_string = '';
				$actual_string = implode ('&', $post_string);
				$result = wp_remote_post($environment_url, array(
				  'method' => 'POST',
				  'timeout'     => 6,
				  'redirection' => 5,
				  'body' => $actual_string,
				));
				$lines = preg_split('/\r\n|\r|\n/', $result['body']);
				$response = array();
				foreach ($lines as $line) {
				    $key_value = preg_split('/=/', $line, 2);
				    if (count($key_value) > 1) {
				        $response[trim($key_value[0])] = trim($key_value[1]);
				    }
				}

				$accepted = array(8, 81, 85); // OK

				$status = preg_replace('/[^a-zA-Z0-9\s]/', '', $response['STATUS']);
				$error = preg_replace('/[^a-zA-Z0-9\s]/', '', $response['NCERROR']);

				if (!is_wp_error($result) && $result['response']['code'] >= 200 && $result['response']['code'] < 300) {

				  if (in_array($status, $accepted)) {
						$order->add_order_note( __('Refund successful', 'ag-sagepay-direct') . '<br />' .
							__('Refund Amount: ', 'ag-sagepay-direct') . $amount . '<br />' .
							__('Refund Reason: ', 'ag-sagepay-direct') . $reason . '<br />' .
							__('ePDQ Status: ', 'ag-sagepay-direct') . get_epdq_status_code($status) . ' '
					);

					return true;


				  } else {
				    $order->add_order_note( __('Refund failed', 'ag-sagepay-direct') . '<br />' . __('ePDQ Status: ', 'ag-sagepay-direct') . get_epdq_status_code($status) . '<br />' );
				  }

				}


	}

} // End of ag_epdq_checkout

/*-----------------------------------------------------------------------------------*/
/*	Getting XML to Array
/*-----------------------------------------------------------------------------------*/
// Get xml responce
function XMLtoArray($XML) {

		$xml_parser = xml_parser_create();
		xml_parse_into_struct($xml_parser, $XML, $vals);
		xml_parser_free($xml_parser);
		// wyznaczamy tablice z powtarzajacymi sie tagami na tym samym poziomie
		$_tmp='';
		foreach ($vals as $xml_elem) {
				$x_tag=$xml_elem['tag'];
				$x_level=$xml_elem['level'];
				$x_type=$xml_elem['type'];
				if ($x_level!=1 && $x_type == 'close') {
						if (isset($multi_key[$x_tag][$x_level]))
								$multi_key[$x_tag][$x_level]=1;
						else
								$multi_key[$x_tag][$x_level]=0;
				}
				if ($x_level!=1 && $x_type == 'complete') {
						if ($_tmp==$x_tag)
								$multi_key[$x_tag][$x_level]=1;
						$_tmp=$x_tag;
				}
		}
		// jedziemy po tablicy
		foreach ($vals as $xml_elem) {
				$x_tag=$xml_elem['tag'];
				$x_level=$xml_elem['level'];
				$x_type=$xml_elem['type'];
				if ($x_type == 'open')
						$level[$x_level] = $x_tag;
				$start_level = 1;
				$php_stmt = '$xml_array';
				if ($x_type=='close' && $x_level!=1)
						$multi_key[$x_tag][$x_level]++;
				while ($start_level < $x_level) {
						$php_stmt .= '[$level['.$start_level.']]';
						if (isset($multi_key[$level[$start_level]][$start_level]) && $multi_key[$level[$start_level]][$start_level])
								$php_stmt .= '['.($multi_key[$level[$start_level]][$start_level]-1).']';
						$start_level++;
				}
				$add='';
				if (isset($multi_key[$x_tag][$x_level]) && $multi_key[$x_tag][$x_level] && ($x_type=='open' || $x_type=='complete')) {
						if (!isset($multi_key2[$x_tag][$x_level]))
								$multi_key2[$x_tag][$x_level]=0;
						else
								$multi_key2[$x_tag][$x_level]++;
						$add='['.$multi_key2[$x_tag][$x_level].']';
				}
				if (isset($xml_elem['value']) && trim($xml_elem['value'])!='' && !array_key_exists('attributes', $xml_elem)) {
						if ($x_type == 'open')
								$php_stmt_main=$php_stmt.'[$x_type]'.$add.'[\'content\'] = $xml_elem[\'value\'];';
						else
								$php_stmt_main=$php_stmt.'[$x_tag]'.$add.' = $xml_elem[\'value\'];';
						eval($php_stmt_main);
				}
				if (array_key_exists('attributes', $xml_elem)) {
						if (isset($xml_elem['value'])) {
								$php_stmt_main=$php_stmt.'[$x_tag]'.$add.'[\'content\'] = $xml_elem[\'value\'];';
								eval($php_stmt_main);
						}
						foreach ($xml_elem['attributes'] as $key=>$value) {
								$php_stmt_att=$php_stmt.'[$x_tag]'.$add.'[$key] = $value;';
								eval($php_stmt_att);
						}
				}
		}
		return $xml_array;
}

/*-----------------------------------------------------------------------------------*/
/*	Status codes
/*-----------------------------------------------------------------------------------*/
function get_epdq_status_code($code){
		if( $code == '' )
			return;
		$codes =  array(
			0=>'Incomplete or invalid',
			1=>'Cancelled by client',
			2=>'Authorisation refused',
			4=>'Order stored',
			40=>'Stored waiting external result',
			41=>'Waiting client payment',
			46=>'Waiting for identification',
			5=>'Authorised',
			51=>'Authorisation waiting',
			52=>'Authorisation not known',
			55=>'Standby',
			56=>'OK with scheduled payments',
			57=>'Not OK with scheduled payments',
			59=>'Author. to get manually',
			6=>'Authorised and canceled',
			61=>'Author. deletion waiting',
			62=>'Author. deletion uncertain',
			63=>'Author. deletion refused',
			64=>'Authorised and cancelled',
			7=>'Payment deleted',
			71=>'Payment deletion pending',
			72=>'Payment deletion uncertain',
			73=>'Payment deletion refused',
			74=>'Payment deleted (not accepted)',
			75=>'Deletion processed by merchant',
			8=>'Refund',
			81=>'Refund pending',
			82=>'Refund uncertain',
			83=>'Refund refused',
			84=>'Payment declined by the acquirer (will be debited)',
			85=>'Refund processed by merchant',
			9=>'Payment requested',
			91=>'Payment processing',
			92=>'Payment uncertain',
			93=>'Payment refused',
			94=>'Refund declined by the acquirer',
			95=>'Payment processed by merchant',
			96=>'Refund reversed',
			97=>'Being processed - intermediate technical status',
			98=>'Being processed - intermediate technical status',
			99=>'Being processed - intermediate technical status'
		);
		if (isset($codes[$code]))
			return $codes[$code];
		else
			return 'Unknown';
	}


/*-----------------------------------------------------------------------------------*/
/*	Error codes
/*-----------------------------------------------------------------------------------*/
function get_epdq_ncerror($code) {
		if ($code == '')
			return;

		$code = (int) $code;
		$ncerorr_list=array(

			20001001=>'Authorisation failed. Please retry.',
			20001002=>'Authorisation failed. Please retry.',
			20001003=>'Authorisation failed. Please retry.',
			20001004=>'Authorisation failed. Please retry.',
			20001005=>'Authorisation failed. Please retry.',
			20001006=>'Authorisation failed. Please retry.',
			20001007=>'Authorisation failed. Please retry.',
			20001008=>'Authorisation failed. Please retry.',
			20001009=>'Authorisation failed. Please retry.',
			20001010=>'Authorisation failed. Please retry.',
			30001999=>'Our payment system is currently under maintenance, please try later.',
			50001005=>'Expiry date error.',
			50001007=>'Requested operation code not permitted.',
			50001008=>'Invalid time limit value.',
			50001010=>'Invalid input date format.',
			50001013=>'Unable to parse socket input stream.',
			50001014=>'Error in parsing stream content.',
			50001015=>'Currency error.',
			50001016=>'Transaction still posted at end of wait.',
			50001017=>'Sync value not compatible with delay value.',
			50001019=>'Duplicate of a pre-existing transaction.',
			50001020=>'Acceptance code required for transaction.',
			50001024=>'Maintenance acquirer differs from original transaction acquirer.',
			50001025=>'Maintenance merchant differs from original transaction merchant.',
			50001028=>'Maintenance operation not appropriate for original transaction.',
			50001031=>'Host application unknown for the transaction.',
			50001032=>'Unable to perform requested operation with requested currency.',
			50001033=>'Maintenance card number differs from original transaction card number.',
			50001034=>'Operation code not permitted.',
			50001035=>'Exception occurred in socket input stream processing.',
			50001036=>'Card length does not correspond to an acceptable value for the brand.',
			50001036=>'Card length does not correspond to an acceptable value for the brand.',
			50001068=>'A technical problem has occurred. Please contact the helpdesk.',
			50001069=>'Invalid check for CardID and Brand.',
			50001070=>'A technical problem has occurred. Please contact the helpdesk.',
			50001116=>'Unknown origin IP.',
			50001117=>'No origin IP detected.',
			50001118=>'Merchant configuration problem. Please contact support.',
			10001001=>'Communication failure.',
			10001002=>'Communication failure.',
			10001003=>'Communication failure.',
			10001004=>'Communication failure.',
			10001005=>'Communication failure.',
			10001016=>'Waiting for acquirer feedback.',
			10001018=>'3XCB pending transaction awaiting for Final status.',
			20001001=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001002=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the status of the transaction within one working day. Please check the status later.',
			20001003=>'We received an unknown status for the transaction. We will contact your acquirer and update the status of the transaction. Please check the status later.',
			20001004=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001005=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001006=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001007=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001008=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001009=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001010=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001101=>'A technical problem has occurred. Please contact the helpdesk.',
			20001104=>'We received an unknown status for the transaction. We will contact your acquirer and update the status of the transaction. Please check the status later.',
			20001105=>'We have received an unknown status for the transaction. We shall contact your acquirer and update the transaction status within one working day. Please check the status later.',
			20001111=>'A technical problem has occurred. Please contact the helpdesk.',
			20001998=>'We received an unknown status for the transaction. We will contact your acquirer and update the status of the transaction. Please check the status later.',
			20002001=>'Bank response origin cannot be checked.',
			20002002=>'Beneficiary account number has been modified during processing.',
			20002003=>'Amount has been modified during processing.',
			20002004=>'Currency has been modified during processing.',
			20002005=>'No feedback detected from the bank server.',
			30001001=>'Payment refused by the financial institution.',
			30001002=>'Duplicate request.',
			30001010=>'A technical problem has occurred. Please contact the helpdesk.',
			30001011=>'A technical problem has occurred. Please contact the helpdesk.',
			30001012=>'Card blacklisted - Contact acquirer.',
			30001015=>'There has been a connection error to the receiving bank. Please try later or choose another payment method.',
			30001016=>'Transmission failure and/or Database error. The transaction could not be properly initialised in the send process (db access failures, etc.).',
			30001051=>'A technical problem has occurred. Please contact the helpdesk.',
			30001054=>'A technical problem has occurred. Please contact the helpdesk.',
			30001056=>'Your merchant\'s acquirer is temporarily unavailable, please try later or choose another payment method.',
			30001057=>'There has been a connection error to the receiving bank. Please try later or choose another payment method.',
			30001058=>'There has been a connection error to the receiving bank. Please try later or choose another payment method.',
			30001060=>'Acquirer has i`ndicated a failure during payment processing.',
			30001070=>'RATEPAY Invalid Response Type (Failure).',
			30001071=>'RATEPAY Missing Mandatory status code field (failure).',
			30001072=>'RATEPAY Missing Mandatory Result code field (failure).',
			30001073=>'RATEPAY Response parsing Failed.',
			30001090=>'CVC check required by front end and returned invalid by acquirer.',
			30001091=>'Postcode check required by front end and returned invalid by acquirer.',
			30001092=>'Address check required by frontend and returned as invalid by acquirer.',
			30001100=>'Unauthorised customer country.',
			30001101=>'IP country differs from card country.',
			30001102=>'Number of different countries too high.',
			30001103=>'unauthorised card country.',
			30001104=>'unauthorised IP address country.',
			30001105=>'Anonymous proxy.',
			30001110=>'If the problem persists, please contact Support or go to paysafecard\'s card balance page (https://customer.cc.at.paysafecard.com/psccustomer/GetWelcomePanelServlet?language=en), to see when the amount reserved on your card will be available again.',
			30001120=>'IP address on merchant\'s blacklist.',
			30001130=>'BIN on merchant\'s blacklist.',
			30001131=>'Wrong BIN for 3xCB.',
			30001140=>'Card on merchant\'s blacklist.',
			30001141=>'E-mail blacklisted.',
			30001142=>'Passenger name blacklisted.',
			30001143=>'Cardholder name blacklisted.',
			30001144=>'Passenger name different from owner name.',
			30001145=>'Time to departure too short.',
			30001149=>'Card Configured in Card Supplier Limit for another relation (CSL).',
			30001150=>'Card not configured in the system for this customer (CSL).',
			30001151=>'REF1 not allowed for this relationship (Contract number).',
			30001152=>'Card/Supplier Amount limit reached (CSL).',
			30001153=>'Card not permitted for this supplier (Date out of contractual limits).',
			30001154=>'You have reached the permitted usage limit.',
			30001155=>'You have reached the permitted usage limit.',
			30001156=>'You have reached the permitted usage limit.',
			30001157=>'Unauthorised IP country for itinerary.',
			30001158=>'e-mail usage limit reached.',
			30001159=>'Unauthorised card country/IP country combination.',
			30001160=>'Postcode in high-risk group.',
			30001161=>'generic blacklist match.',
			30001162=>'Invoicing Address is a PO Box.',
			30001163=>'Airport in high-risk group.',
			30001164=>'Shipping Method in high-risk group.',
			30001165=>'Shipping Method Details in high-risk group.',
			30001166=>'Product Category in high-risk group.',
			30001167=>'Subbrand in high-risk group.',
			30001168=>'Issuer Number in high-risk group.',
			30001169=>'Time to delivery too short.',
			30001180=>'maximum scoring reached.',
			30001997=>'Authorisation cancelled by simulator.',
			30001998=>'A technical problem has occurred. Please try again.',
			30001999=>'There has been a connection error with the receiving bank. Please try later or choose another payment method.',
			30002001=>'Payment refused by the financial institution.',
			30002001=>'Payment refused by the financial institution.',
			30021001=>'Please call the acquirer support call number.',
			30022001=>'Payment must be approved by the acquirer prior to execution.',
			30031001=>'Invalid merchant number.',
			30041001=>'Retain card.',
			30051001=>'Authorisation declined.',
			30051002=>'Voor vragen over uw afwijzing kunt u contact opnemen met de @STARTURL@http://www.afterpay.nl/consument-contact@TXTURL@Klantenservice van AfterPay@ENDURL@.',
			30051009=>'It is possible that you may not have completed all the required information (correctly) during the order process.',
			30051010=>'because your age is under 18. For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=2@TXTURL@website of AfterPay@ENDURL@.',
			30051011=>'because your address could not be validated. For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=3@TXTURL@website of AfterPay@ENDURL@.',
			30051012=>'because your emailadres is invalid. For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=4@TXTURL@website van AfterPay@ENDURL@.',
			30051013=>'because the order amount extends the limit for first time AfterPay users. For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=5@TXTURL@website of AfterPay@ENDURL@.',
			30051014=>'because there are currently too many outstanding payments at AfterPay. For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=6@TXTURL@website of AfterPay@ENDURL@.',
			30051015=>'because your chamber of commerce number could not be validated. For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=7@TXTURL@website of AfterPay@ENDURL@.',
			30051016=>'because the order amount is too low. For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=8@TXTURL@website of AfterPay@ENDURL@.',
			30051017=>'For more information please visit @STARTURL@http://www.afterpay.nl/rej.php?LAN=$LANGUAGE$&TEMPLATE=1@TXTURL@website of AfterPay@ENDURL@.',
			30071001=>'Retain card - special conditions.',
			30121001=>'Invalid transaction.',
			30131001=>'Invalid amount.',
			30131002=>'You have reached the permitted limit.',
			30141001=>'Invalid card number.',
			30151001=>'Unknown acquiring institution.',
			30171001=>'Payment method cancelled by the customer.',
			30171002=>'The maximum time allowed has elapsed.',
			30191001=>'Please try again later.',
			30201001=>'A technical problem has occurred. Please contact the helpdesk.',
			30301001=>'Invalid format.',
			30311001=>'Unknown acquirer ID.',
			30331001=>'Card expired.',
			30341001=>'Suspicion of fraud.',
			30341001=>'Suspicion of fraud.',
			30341002=>'Suspicion of fraud (3rdMan).',
			30341003=>'Suspicion of fraud (Perseuss).',
			30341004=>'Suspicion of fraud (ETHOCA).',
			30341005=>'Suspicion of fraud (Expert).',
			30381001=>'A technical problem has occurred. Please contact the helpdesk.',
			30401001=>'Invalid function.',
			30411001=>'Lost card.',
			30431001=>'Stolen card. Pick up.',
			30511001=>'Insufficient funds.',
			30521001=>'No Authorisation. Please contact your card issuer.',
			30541001=>'Card expired.',
			30551001=>'Invalid PIN.',
			30561001=>'Card not in authoriser\'s database.',
			30571001=>'Transaction not permitted on card.',
			30581001=>'Transaction not permitted on this terminal.',
			30591001=>'Suspicion of fraud.',
			30601001=>'The merchant should contact the acquirer.',
			30611001=>'Amount exceeds card limit.',
			30621001=>'Restricted card.',
			30631001=>'Security policy not respected.',
			30641001=>'Amount changed from ref. transaction.',
			30681001=>'The maximum allowed time has elapsed.',
			30751001=>'Incorrect PIN entered too many times.',
			30761001=>'Already disputed by cardholder.',
			30771001=>'PIN entry required.',
			30811001=>'Message flow error.',
			30821001=>'Authorisation centre unavailable.',
			30831001=>'Authorisation centre unavailable.',
			30901001=>'Temporary system shutdown.',
			30911001=>'Acquirer unavailable.',
			30921001=>'Invalid card type for acquirer.',
			30941001=>'Duplicate transaction.',
			30961001=>'Processing temporarily not possible.',
			30971001=>'A technical problem has occurred. Please contact the helpdesk.',
			30981001=>'A technical problem has occurred. Please contact the helpdesk.',
			31011001=>'Unknown acceptance code.',
			31021001=>'Invalid currency.',
			31031001=>'Acceptance code missing.',
			31041001=>'Inactive card.',
			31051001=>'Merchant not active.',
			31061001=>'Invalid expiry date.',
			31071001=>'Interrupted host communication.',
			31081001=>'Card refused.',
			31091001=>'Invalid password.',
			31101001=>'Plafond transaction (majoré du bonus) dépassé.',
			31111001=>'Plafond mensuel (majoré du bonus) dépassé.',
			31121001=>'Plafond centre de facturation dépassé.',
			31131001=>'Plafond entreprise dépassé.',
			31141001=>'Code MCC du fournisseur non autorisé pour la carte.',
			31151001=>'Numéro SIRET du fournisseur non autorisé pour la carte.',
			31161001=>'This is not a valid online bank account.',
			32001004=>'A technical problem has occurred. Please try again.',
			32001105=>'A technical problem has occurred. Please contact the helpdesk.',
			34011001=>'Bezahlung mit RatePAY nicht möglich.',
			39991001=>'A technical problem has occurred. Please contact your acquirer\'s helpdesk.',
			40001001=>'A technical problem has occurred. Please try again.',
			40001002=>'A technical problem has occurred. Please try again.',
			40001003=>'A technical problem has occurred. Please try again.',
			40001004=>'A technical problem has occurred. Please try again.',
			40001005=>'A technical problem has occurred. Please try again.',
			40001006=>'A technical problem has occurred. Please try again.',
			40001007=>'A technical problem has occurred. Please try again.',
			40001008=>'A technical problem has occurred. Please try again.',
			40001009=>'A technical problem has occurred. Please try again.',
			40001010=>'A technical problem has occurred. Please try again.',
			40001011=>'A technical problem has occurred. Please contact the helpdesk.',
			40001012=>'There has been a connection error with the receiving bank. Please try later or choose another payment method.',
			40001013=>'A technical problem has occurred. Please contact the helpdesk.',
			40001016=>'A technical problem has occurred. Please contact the helpdesk.',
			40001018=>'A technical problem has occurred. Please try again.',
			40001019=>'Sorry, an error has occurred during processing. Please retry the transaction (using the Back button of the browser). If the problem persists, contact your merchant\'s helpdesk.',
			40001020=>' Sorry, an error occurred during processing. Please retry the operation (using the Back button of the browser). If the problem persists, please contact your merchant\'s helpdesk.',
			40001050=>'A technical problem has occurred. Please contact the helpdesk.',
			40001133=>'Authentication failed. Incorrect signature for your bank\'s access control server.',
			40001134=>'Authentication failed. Please retry or cancel.',
			40001135=>'Authentication temporarily unavailable. Please retry or cancel.',
			40001136=>'Technical problem with your browser. Please retry or cancel.',
			40001137=>'Your bank is temporarily unavailable. Please try again later or choose another payment method.',
			40001998=>'Temporary technical problem. Please retry later.',
			50001001=>'Unknown card type.',
			50001002=>'Card number format check failed for given card number.',
			50001003=>'Merchant data error.',
			50001004=>'Merchant identification missing.',
			50001005=>'Expiry date error.',
			50001006=>'Amount is not a number.',
			50001007=>'A technical has problem occurred. Please contact the helpdesk.',
			50001008=>'A technical has problem occurred. Please contact the helpdesk.',
			50001009=>'A technical has problem occurred. Please contact the helpdesk.',
			50001010=>'A technical has problem occurred. Please contact the helpdesk.',
			50001011=>'Brand not supported for that merchant.',
			50001012=>'A technical has problem occurred. Please contact the helpdesk.',
			50001013=>'A technical has problem occurred. Please contact the helpdesk.',
			50001014=>'A technical has problem occurred. Please contact the helpdesk.',
			50001015=>'Invalid currency code.',
			50001016=>'A technical has problem occurred. Please contact the helpdesk.',
			50001017=>'A technical has problem occurred. Please contact the helpdesk.',
			50001018=>'A technical has problem occurred. Please contact the helpdesk.',
			50001019=>'A technical has problem occurred. Please contact the helpdesk.',
			50001020=>'A technical has problem occurred. Please contact the helpdesk.',
			50001021=>'A technical has problem occurred. Please contact the helpdesk.',
			50001022=>'A technical has problem occurred. Please contact the helpdesk.',
			50001023=>'A technical has problem occurred. Please contact the helpdesk.',
			50001024=>'A technical has problem occurred. Please contact the helpdesk.',
			50001025=>'A technical has problem occurred. Please contact the helpdesk.',
			50001026=>'A technical has problem occurred. Please contact the helpdesk.',
			50001027=>'A technical has problem occurred. Please contact the helpdesk.',
			50001028=>'A technical has problem occurred. Please contact the helpdesk.',
			50001029=>'A technical has problem occurred. Please contact the helpdesk.',
			50001030=>'A technical has problem occurred. Please contact the helpdesk.',
			50001031=>'A technical has problem occurred. Please contact the helpdesk.',
			50001032=>'A technical has problem occurred. Please contact the helpdesk.',
			50001033=>'A technical has problem occurred. Please contact the helpdesk.',
			50001034=>'A technical has problem occurred. Please contact the helpdesk.',
			50001035=>'A technical problem has occurred. Please contact the helpdesk.',
			50001036=>'Incorrect card length for the brand.',
			50001037=>'Purchasing card number for a standard merchant.',
			50001038=>'You should use a purchasing card for this transaction.',
			50001039=>'Details sent for a non-purchasing card merchant. Please contact the helpdesk.',
			50001040=>'Details not sent for a purchasing card transaction. Please contact the helpdesk.',
			50001041=>'Payment detail validation failed.',
			50001042=>'Sum of given transaction amounts (tax, discount, delivery, net, etc.) does not match total.',
			50001043=>'A technical problem has occurred. Please contact the helpdesk.',
			50001044=>'No acquirer configured for this operation.',
			50001045=>'No UID configured for this operation.',
			50001046=>'Operation not permitted for the merchant.',
			50001047=>'A technical problem has occurred. Please contact the helpdesk.',
			50001048=>'A technical problem has occurred. Please contact the helpdesk.',
			50001049=>'A technical problem has occurred. Please contact the helpdesk.',
			50001050=>'A technical problem has occurred. Please contact the helpdesk.',
			50001051=>'A technical problem has occurred. Please contact the helpdesk.',
			50001052=>'A technical problem has occurred. Please contact the helpdesk.',
			50001053=>'A technical problem has occurred. Please contact the helpdesk.',
			50001054=>'Card number incorrect or incompatible.',
			50001055=>'A technical problem has occurred. Please contact the helpdesk.',
			50001056=>'A technical problem has occurred. Please contact the helpdesk.',
			50001057=>'A technical problem has occurred. Please contact the helpdesk.',
			50001058=>'A technical problem has occurred. Please contact the helpdesk.',
			50001059=>'A technical problem has occurred. Please contact the helpdesk.',
			50001060=>'A technical problem has occurred. Please contact the helpdesk.',
			50001061=>'A technical problem has occurred. Please contact the helpdesk.',
			50001062=>'A technical problem has occurred. Please contact the helpdesk.',
			50001063=>'Card Issue Number does not correspond to range or is not present.',
			50001064=>'Start Date invalid or not present.',
			50001066=>'Invalid CVC code format.',
			50001067=>'The merchant is not registered for 3D-Secure.',
			50001068=>'Invalid card number or account number (PAN).',
			50001069=>'Invalid CardID and Brand match.',
			50001070=>'The ECI value is either not supported or conflicts with other transaction data.',
			50001071=>'Incomplete TRN demat.',
			50001072=>'Incomplete PAY demat.',
			50001073=>'No demat APP.',
			50001074=>'Authorisation period expired.',
			50001075=>'VERRes was an error message.',
			50001076=>'DCP amount greater than authorisation amount.',
			50001077=>'Details negative amount.',
			50001078=>'Details negative quantity.',
			50001079=>'Could not decode/decompress received PARes (3D-Secure).',
			50001080=>'Received PARes was an error message from ACS (3D-Secure).',
			50001081=>'Received PARes format was invalid according to the 3DS specifications (3D-Secure).',
			50001082=>'PAReq/PARes reconciliation failure (3D-Secure).',
			50001084=>'Maximum amount reached.',
			50001087=>'This transaction requires authentication. Please check with your bank.',
			50001090=>'CVC missing at input, but CVC check requested.',
			50001091=>'Postcode missing at input, but postcode check requested.',
			50001092=>'Address missing at input, but Address check requested.',
			50001093=>'Partial capture not allowed.',
			50001095=>'Invalid date of birth.',
			50001096=>'Invalid commodity code.',
			50001097=>'The requested currency and brand are incompatible.',
			50001111=>'Data validation error.',
			50001113=>'This order has already been processed.',
			50001114=>'Error in accessing the pre-payment check page.',
			50001115=>'Request not received in secure mode.',
			50001116=>'Unknown IP address origin.',
			50001117=>'No IP address origin.',
			50001118=>'PSPID not found or incorrect.',
			50001119=>'Password incorrect or disabled due to number of errors.',
			50001120=>'Invalid currency.',
			50001121=>'Invalid number of decimals for the currency.',
			50001122=>'Currency not accepted by the merchant.',
			50001123=>'Card type not active.',
			50001124=>'Number of lines doesn\'t match the number of payments.',
			50001125=>'Format validation error.',
			50001126=>'Overflow in data capture requests for the original order.',
			50001127=>'Incorrect original order status.',
			50001128=>'missing authorisation code for unauthorised order.',
			50001129=>'Overflow in refunds requests.',
			50001130=>'Original order access error.',
			50001131=>'Original history item access error.',
			50001132=>'The selected Catalogue is empty.',
			50001133=>'Duplicate request.',
			50001134=>'Authentication failed. Please retry or cancel.',
			50001135=>'Authentication temporarily unavailable. Please retry or cancel.',
			50001136=>'Technical problem with your browser. Please retry or cancel.',
			50001137=>'Your bank is temporarily unavailable. Please try again later or choose another payment method.',
			50001150=>'Fraud Detection: technical error (invalid IP).',
			50001151=>'Fraud detection: technical error (IPCTY unknown or error).',
			50001152=>'Fraud detection: technical error (CCCTY unknown or error).',
			50001153=>'Overflow in redo-authorisation requests.',
			50001170=>'Dynamic BIN check failed.',
			50001171=>'Dynamic country check failed.',
			50001172=>'Amadeus signature error.',
			50001174=>'Cardholder Name is too long.',
			50001175=>'Name contains invalid characters.',
			50001176=>'Card number is too long.',
			50001177=>'Card number contains non-numeric info.',
			50001178=>'Card Number Empty.',
			50001179=>'CVC too long.',
			50001180=>'CVC contains non-numeric info.',
			50001181=>'Expiry date contains non-numeric info.',
			50001182=>'Invalid expiry month.',
			50001183=>'Expiry date must be in the future.',
			50001184=>'SHA Mismatch.',
			50001186=>'Operation not permitted.',
			50001187=>'Operation not permitted.',
			50001205=>'Missing mandatory fields in invoicing address.',
			50001206=>'Missing mandatory date of birth field.',
			50001207=>'Missing required shopping basket details.',
			50001208=>'Missing social security number.',
			50001209=>'Invalid country code.',
			50001210=>'Missing annual salary.',
			50001211=>'Missing gender.',
			50001212=>'Missing e-mail.',
			50001213=>'Missing IP address.',
			50001214=>'Missing part-payment campaign ID.',
			50001215=>'Missing invoice number.',
			50001216=>'The alias must be different to the card number.',
			50001217=>'Invalid details for shopping basket calculation.',
			50001218=>'No Refunds allowed.',
			50001220=>'Invalid format of phone number.',
			50001221=>'Invalid ZIP format.',
			50001222=>'Firstname or/and lastname missing.',
			50001223=>'Firstname and/or lastname format invalid.',
			50001224=>'The phone number is missing.',
			50001225=>'Invalid email format.',
			50001300=>'Wrong brand/payment method.',
			50001301=>'Wrong account number format.',
			50001302=>'RFP operation code is only permitted with scheduled payments.',
			50001303=>'RFP operation code not permitted for a Disputed payment.',
			50001304=>'RFP operation code not permitted - Unpaid amounts.',
			55555555=>'An error occurred.',
			60000001=>'account number unknown.',
			60000003=>'not credited dd-mm-yy.',
			60000005=>'name/number do not match.',
			60000007=>'account number blocked.',
			60000008=>'specific direct debit block.',
			60000009=>'account number WKA.',
			60000010=>'administrative reason.',
			60000011=>'account number expired.',
			60000012=>'no direct debit authorisation.',
			60000013=>'debit not approved.',
			60000014=>'double payment.',
			60000018=>'name/address/city not entered.',
			60001001=>'no original direct debit for revocation.',
			60001002=>'payer\'s account number format error.',
			60001004=>'payer\'s account at different bank.',
			60001005=>'payee\'s account at different bank.',
			60001006=>'payee\'s account number format error.',
			60001007=>'payer\'s account number blocked.',
			60001008=>'payer\'s account number expired.',
			60001009=>'payee\'s account number expired.',
			60001010=>'direct debit not possible.',
			60001011=>'creditor payment not possible.',
			60001012=>'payer\'s account number unknown WKA-number.',
			60001013=>'payee\'s account number unknown WKA-number.',
			60001014=>'WKA transaction not permitted.',
			60001015=>'revocation period expired.',
			60001017=>'incorrect revocation reason.',
			60001018=>'original run number not numeric.',
			60001019=>'payment ID incorrect.',
			60001020=>'amount not numeric.',
			60001021=>'zero amount not permitted.',
			60001022=>'negative amount not permitted.',
			60001023=>'payer and payee giro account number.',
			60001025=>'processing code (verwerkingscode) incorrect.',
			60001028=>'revocation not permitted.',
			60001029=>'guaranteed direct debit on giro account number.',
			60001030=>'NBC transaction type incorrect.',
			60001031=>'description too long.',
			60001032=>'book account number not issued.',
			60001034=>'book account number incorrect.',
			60001035=>'payer\'s account number not numeric.',
			60001036=>'payer\'s account number not eleven-proof.',
			60001037=>'payer\'s account number not issued.',
			60001039=>'payer\'s account number of DNB/BGC/BLA.',
			60001040=>'payee\'s account number not numeric.',
			60001041=>'payee\'s account number not eleven-proof.',
			60001042=>'payee\'s account number not issued.',
			60001044=>'payee\'s account number unknown.',
			60001050=>'payee\'s name missing.',
			60001051=>'indicate payee\'s bank account number instead of 3102.',
			60001052=>'no direct debit contract.',
			60001053=>'amount beyond limits.',
			60001054=>'selective direct debit block.',
			60001055=>'original run number unknown.',
			60001057=>'payer\'s name missing.',
			60001058=>'payee\'s account number missing.',
			60001059=>'restore not permitted.',
			60001060=>'bank\'s reference (navraaggegeven) missing.',
			60001061=>'BEC/GBK number incorrect.',
			60001062=>'BEC/GBK code incorrect.',
			60001087=>'book account number not numeric.',
			60001090=>'cancelled on request.',
			60001091=>'cancellation order executed.',
			60001092=>'cancelled instead of ended.',
			60001093=>'book account number is a shortened account number.',
			60001094=>'instructing party and payer account numbers do not match.',
			60001095=>'payee unknown GBK acceptor.',
			60001097=>'instructing party and payee account numbers do not match.',
			60001099=>'clearing not permitted.',
			60001101=>'payer\'s account number has no spaces.',
			60001102=>'PAN length not numeric.',
			60001103=>'PAN length outside limits.',
			60001104=>'track number not numeric.',
			60001105=>'track number not valid.',
			60001106=>'PAN sequence number not numeric.',
			60001107=>'domestic PAN not numeric.',
			60001108=>'domestic PAN not eleven-proof.',
			60001109=>'domestic PAN not issued.',
			60001110=>'foreign PAN not numeric.',
			60001111=>'card validity date not numeric.',
			60001112=>'book period number (boekperiodenr) not numeric.',
			60001113=>'transaction number not numeric.',
			60001114=>'transaction time not numeric.',
			60001115=>'invalid transaction time.',
			60001116=>'transaction date not numeric.',
			60001117=>'invalid transaction date.',
			60001118=>'STAN not numeric.',
			60001119=>'instructing party\'s name missing.',
			60001120=>'foreign amount (bedrag-vv) not numeric.',
			60001122=>'rate (verrekenkoers) not numeric.',
			60001125=>'number of decimals (aantaldecimalen) incorrect.',
			60001126=>'tariff (tarifering) not B/O/S.',
			60001127=>'domestic costs (kostenbinnenland) not numeric.',
			60001128=>'domestic costs (kostenbinnenland) not higher than zero.',
			60001129=>'foreign costs (kostenbuitenland) not numeric.',
			60001130=>'foreign costs (kostenbuitenland) not higher than zero.',
			60001131=>'domestic costs (kostenbinnenland) not zero.',
			60001132=>'foreign costs (kostenbuitenland) not zero.',
			60001134=>'Euro record not completed.',
			60001135=>'Customer currency incorrect.',
			60001136=>'NLG amount not numeric.',
			60001137=>'NLG amount not higher than zero.',
			60001138=>'NLG amount not equal to Amount.',
			60001139=>'NLG amount incorrectly converted.',
			60001140=>'EUR amount not numeric.',
			60001141=>'EUR amount not greater than zero.',
			60001142=>'EUR amount not equal to Amount.',
			60001143=>'EUR amount incorrectly converted.',
			60001144=>'Customer currency not NLG.',
			60001145=>'rate euro-vv (Koerseuro-vv) not numeric.',
			60001146=>'comma rate euro-vv (Kommakoerseuro-vv) incorrect.',
			60001147=>'invalid acceptgiro distributor.',
			60001148=>'Original run number and/or BRN missing.',
			60001149=>'Amount/Account number/ BRN different.',
			60001150=>'Direct debit already revoked/restored.',
			60001151=>'Direct debit already reversed/revoked/restored.',
			60001153=>'Payer\'s account number not know.',
		);


		if( isset($ncerorr_list[$code] ))
			return $ncerorr_list[$code];
		else
			return 'Unknown';
	}
