<?php
# @Author: Aaron Bowie
# @Date:   Friday, June 16th 2017, 11:11:20 pm
# @Email:  support@weareag.co.uk
# @Filename: epdq.php
# @Last modified by:   Aaron Bowie
# @Last modified time: Wednesday, November 15th 2017, 10:52:01 am

/*
Plugin Name: AG Barclaycard ePDQ Direct Link Payment Gateway for WooCommerce
Plugin URI: https://www.weareag.co.uk/
Description: Extends WooCommerce by Adding the Barclaycard ePDQ Direct Link Payment Gateway.
Version: 1.1.1
Author: Aaron Bowie, We are AG
Author URI: https://www.weareag.co.uk/
*/

defined('ABSPATH') or die("No script kiddies please!");


// Plugin repo bits
$version = '1.1.1'; 
$plugin_id = '1c4s80ZjUSkh0i1i';
$repo_url = 'http://weareag.co.uk/update/';
// defined
if ( ! defined( 'AG_ePDQ_Repo' ) ) {
	define( 'AG_ePDQ_Repo', $repo_url );
}
if ( ! defined( 'plugin_id_ag' ) ) {
	define( 'plugin_id_ag', $plugin_id );
}
if ( ! defined( 'ag_PLUGIN_VER' ) ) {
	define( 'ag_PLUGIN_VER', $version );
}

if ( ! defined( 'AGD_KEY' ) ) {
	define( 'AGD_KEY', '56cdf3f59bdb87.92903929' );
}
if ( ! defined( 'AGD' ) ) {
	define( 'AGD', 'https://www.weareag.co.uk' );
}
if ( ! defined( 'ITEM' ) ) {
	define( 'ITEM', '2784' );
}

// Include our Gateway Class and register Payment Gateway with WooCommerce
add_action( 'plugins_loaded', 'ag_epdq_checkout_init', 0 );
function ag_epdq_checkout_init() {
	// If the parent WC_Payment_Gateway class doesn't exist
	// it means WooCommerce is not installed on the site
	// so do nothing
	if ( ! class_exists( 'WC_Payment_Gateway' ) ) return;

	// If we made it this far, then include our Gateway Class
	include_once( 'epdq-class.php' );

	// Now that we have successfully included our class,
	// Lets add it too WooCommerce
	add_filter( 'woocommerce_payment_gateways', 'ag_epdq_checkout_gateway' );
	function ag_epdq_checkout_gateway( $methods ) {
		$methods[] = 'ag_epdq_checkout';
		return $methods;
	}
}


function rkv_load_update() {
	if ( ! class_exists( 'RKV_Remote_Updater' ) ) {
		include( 'lib/RKV_Remote_Updater.php' );
	}
}

add_action( 'admin_init', 'agd_ePDQ_remote_update' );
function agd_ePDQ_remote_update() {

	// ensure the class exists before running
	if ( ! class_exists( 'RKV_Remote_Updater' ) ) {
		return;
	}

	$updater = new RKV_Remote_Updater( AG_ePDQ_Repo, __FILE__, array(
		'unique'    => plugin_id_ag,
		'version'   => ag_PLUGIN_VER,
	)
	);
}


// Add custom action links
add_filter( 'plugin_action_links_' . plugin_basename( __FILE__ ), 'ag_epdq_checkout_action_links' );
function ag_epdq_checkout_action_links( $links ) {
	$plugin_links = array(
		'<a href="' . admin_url( 'admin.php?page=wc-settings&tab=checkout' ) . '">' . __( 'Settings', 'ag_epdq_checkout' ) . '</a>',
	);
	// Merge our new link with the default ones
	return array_merge( $plugin_links, $links );
}


add_action('admin_menu', 'epdqapi_licence');
function epdqapi_licence() {
    add_menu_page('AG License Activation Menu', 'AG License', 'manage_options', __FILE__, 'AG_licence_page');
}

function AG_licence_page() {
    echo '<div class="wrap">';
    echo '<h2>We are AG Plugin License Management</h2>';
    /*** License activate button was clicked ***/

    if (isset($_REQUEST['AGD_epdqapi_Key_activate_license'])) {
        $license_key = $_REQUEST['AGD_epdqapi_Key'];
        // API query parameters
        $api_params = array(
            'slm_action' => 'slm_activate',
            'secret_key' => AGD_KEY,
            'license_key' => $license_key,
            'registered_domain' => $_SERVER['SERVER_NAME'],
            'item_reference' => urlencode(pro),
        );
        // Send query to the license manager server
        $query = esc_url_raw(add_query_arg($api_params, AGD));
        $response = wp_remote_get($query, array('timeout' => 20, 'sslverify' => false));
        $license_data = json_decode(wp_remote_retrieve_body($response));

        if($license_data->result == 'success'){//Success was returned for the license activation
            update_option('AGD_epdqapi_Key', $license_key);
            update_option('AGD_epdqapi_Key_message', $license_data->message);
            add_option('AGD_epdqapi_Key_message', $license_data->message);
            $fulldate = $response['headers']['expires'];
            $words = explode(" ", $fulldate);
            $date = join(" ", array_slice($words, 0, 4));
            add_option('AGD_epdqapi_Key_expires', $date);
            update_option('AGD_epdqapi_Key_expires', $date);
        } else {
          update_option('AGD_epdqapi_Key_message', $license_data->message);
        }
    }
    /*** End of license activation ***/
    /*** License activate button was clicked ***/
    if (isset($_REQUEST['AGD_epdqapi_Key_deactivate_license'])) {
        $license_key = $_REQUEST['AGD_epdqapi_Key'];
        // API query parameters
        $api_params = array(
            'slm_action' => 'slm_deactivate',
            'secret_key' => AGD_KEY,
            'license_key' => $license_key,
            'registered_domain' => $_SERVER['SERVER_NAME'],
            'item_reference' => urlencode(pro),
        );
        $query = esc_url_raw(add_query_arg($api_params, AGD));
        $response = wp_remote_get($query, array('timeout' => 20, 'sslverify' => false));

        // License data.
        $license_data = json_decode(wp_remote_retrieve_body($response));
        // TODO - Do something with it.
        if($license_data->result == 'success'){//Success was returned for the license activation
            //Remove the licensse key from the options table. It will need to be activated again.
            update_option('AGD_epdqapi_Key', '');
            update_option('AGD_epdqapi_Key_message', $license_data->message);
            echo $response->headers->expires;
        } else {
          update_option('AGD_epdqapi_Key_message', $license_data->message);
        }
    }
?>
<p>Please enter the license key for this product to activate it. You were given a license key when you purchased this item.</p>
<form action="" method="post">
    <table class="form-table">
        <tr>
          <label class="licence_label" for="AGD_epdqapi_Key">ePDQ Direct link Licence Key</label>
<th style="width:10px;"><input class="regular-text" type="password" id="AGD_epdqapi_Key" name="AGD_epdqapi_Key"  value="<?php echo get_option('AGD_epdqapi_Key'); ?>" ><input type="submit" name="AGD_epdqapi_Key_activate_license" value="Activate" class="button-primary activate_license" /><input type="submit" name="AGD_epdqapi_Key_deactivate_license" value="Deactivate" class="button" />
<?php if (get_option('AGD_epdqapi_Key_message') == 'License key activated') { echo '<p style="color: green"><strong>'. get_option('AGD_epdqapi_Key_message'). '</strong></p>'; } elseif (get_option('AGD_epdqapi_Key_message') == 'Your License key has expired') {
  echo '<p style="color: red"><strong>'. get_option('AGD_epdqapi_Key_message'). ' to continue getting support and updates make sure you <a href="">update your licence here</a> now.</strong></p>';
} else { echo '<p style="color: red"><strong>'. get_option('AGD_epdqapi_Key_message'). '</strong></p>'; } ?></th>
        </tr>
    </table>
</form>
<h4>The plugin settings can be found <a href="<?php echo site_url(); ?>/wp-admin/admin.php?page=wc-settings&tab=checkout&section=ag_epdq_checkout">here</a>.</h4>
<p>
  When moving plugin from one site to another (from development server to live site) you <strong>MUST</strong> deactivate your plugin first on development before you move it over.
</p>
<?php
    echo '</div>';
}
